# -*- coding: utf-8 -*-
"""
    sphinxnotes
    ~~~~~~~~~~~

    This package is a namespace package that contains all extensions
    distributed in the ``sphinx-notes`` distribution.

    :copyright: Copyright 2020 by the Shengyu Zhang.
"""

__import__('pkg_resources').declare_namespace(__name__)

